/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "unnamed.h"

typedef struct
{
  ip_range_t range;
  address_t ip;
} catchall_t;

typedef struct
{
  catchall_t *catchall;
  int catchall_count;
  int trace;
  uint32_t ttl;
} module_data_t;

static void term(int module_index, state_t *state)
{
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  free(module->catchall);
  free(module);
  state->module_data[module_index] = NULL;
}

static int init(int module_index, state_t *state)
{
  int count;
  buffer_t buffer;
  module_data_t *module = (module_data_t*)calloc(1, sizeof(*module));
  if(module == NULL) return -1;
  state->module_data[module_index] = module;

  buffer_init(&buffer);
  module->trace = state->config_get_int(state, "catchall.trace", state->trace);

  module->ttl = (uint32_t)state->config_get_int(state, "catchall.ttl", DEFAULT_TTL);
  count = state->config_maxn(state, "catchall");
  if((module->catchall = (catchall_t*)calloc((size_t)count, sizeof(module->catchall[0]))) == NULL)
  {
    ERROR_IF(module->trace) (void)state->lprintf(TRACE, "calloc failed");
    state->buffer_free(&buffer);
    term(module_index, state);
    return -1;
  }
  for(module->catchall_count = 0; module->catchall_count < count; module->catchall_count++)
  {
    memset(&module->catchall[module->catchall_count], 0, sizeof(module->catchall[0]));
    if(state->buffer_set(&buffer, "catchall.%d.range", module->catchall_count + 1) != 0)
    {
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "buffer_set failed");
      state->buffer_free(&buffer);
      term(module_index, state);
      return -1;
    }
    if(state->config_get_ip_range(state, (char*)buffer.value, &module->catchall[module->catchall_count].range) != 0)
    {
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "config_get_ip_range failed on %s", (char*)buffer.value);
      state->buffer_free(&buffer);
      term(module_index, state);
      return -1;
    }
    if(state->buffer_set(&buffer, "catchall.%d.ip", module->catchall_count + 1) != 0)
    {
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "buffer_set failed");
      state->buffer_free(&buffer);
      term(module_index, state);
      return -1;
    }
    if(state->config_get_ip(state, (char*)buffer.value, &module->catchall[module->catchall_count].ip) != 0)
    {
      const char *ip = state->config_get_const(state, (char*)buffer.value);
      if(ip == NULL)
      {
        ERROR_IF(module->trace) (void)state->lprintf(TRACE, "Error %s is not set", (char*)buffer.value);
      }
      else
      {
        ERROR_IF(module->trace) (void)state->lprintf(TRACE, "Error %s = %s is not valid", (char*)buffer.value, ip);
      }
      state->buffer_free(&buffer);
      term(module_index, state);
      return -1;
    }
  }
  state->buffer_free(&buffer);
  state->config_free_node(state, "catchall");
  return 0;
}

static answer_t answer(int module_index, client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int __attribute__((unused)) lazy)
{
  int i;
  rr_t *rr = NULL;
  char address[IPCHAR_SIZE];
  state_t *state = client->state;
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  if(dclass != CLASS_IN) return ANSWER_UNSATISFIED;

  address[0] = '\0';

  for(i = 0; i < module->catchall_count; i++)
  {
    if(state->is_in_range(&client->from_addr, &module->catchall[i].range) != 0)
    {
      if((module->catchall[i].ip.a.sockaddr.sa_family == AF_INET && type == TYPE_A) || (module->catchall[i].ip.a.sockaddr.sa_family == AF_INET6 && type == TYPE_AAAA))
      {
        if((rr = state->rr_create(name, type, dclass, module->ttl, RCODE_NO_ERROR)) != NULL)
        {
          memcpy(&rr->p.address, &module->catchall[i].ip, sizeof(rr->p.address));
          if(state->add_rr(client, rr_section, &rr) == 0)
          {
            state->free_rr(&rr);
            INFO_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %s", (char*)name->value, state->class_name(dclass), state->type_name(type), state->addr2char(address, &module->catchall[i].ip));
            return ANSWER_SATISFIED;
          }
          state->free_rr(&rr);
        }
      }
    }
  }
  return ANSWER_UNSATISFIED;
}

module_t unnamed_module = { "catchall", NULL, init, answer, NULL, term };
