/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "unnamed.h"

typedef struct
{
  int family;
  ip_range_t range;
} rbl_ip_t;

typedef struct
{
  char *rbl_domain, **rbl_pattern;
  rbl_ip_t *rbl_ip;
  int rbl_pattern_count, rbl_ip_count;
  address_t ip;
  uint32_t ttl;
  int trace;
} module_data_t;

static void term(int module_index, state_t *state)
{
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  free(module->rbl_domain);
  if(module->rbl_pattern != NULL)
  {
    int i;
    for(i = 0; i < module->rbl_pattern_count; i++) free(module->rbl_pattern[i]);
    free(module->rbl_pattern);
  }
  free(module->rbl_ip);
  free(module);
  state->module_data[module_index] = NULL;
}

static int init(int module_index, state_t *state)
{
  int count;
  buffer_t buffer;
  module_data_t *module = (module_data_t*)calloc(1, sizeof(*module));
  if(module == NULL) return -1;
  state->module_data[module_index] = module;

  #define CALLOC_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "calloc failed"); term(module_index, state); state->buffer_free(&buffer); return -1; }
  #define BUFFER_SET_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "buffer_set failed"); term(module_index, state); state->buffer_free(&buffer); return -1; }
  #define CONFIG_GET_IP_RANGE_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "config_get_ip_range failed on %s", (char*)buffer.value); term(module_index, state); state->buffer_free(&buffer); return -1; }
  #define CONFIG_GET_LOWER_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "config_get_lower failed on %s", (char*)buffer.value); term(module_index, state); state->buffer_free(&buffer); return -1; }
  #define CHAR2ADDR_FAILED { ERROR_IF(module->trace) (void)state->lprintf(TRACE, "calloc failed"); term(module_index, state); state->buffer_free(&buffer); return -1; }

  module->trace = state->config_get_int(state, "rbl.trace", TRACE_LEVEL_ERROR);
  module->ttl = (uint32_t)state->config_get_int(state, "rbl.ttl", DEFAULT_TTL);
  buffer_init(&buffer);
  if((module->rbl_domain = state->config_get_lower(state, "rbl.domain")) == NULL) CONFIG_GET_LOWER_FAILED;

  if(state->config_get_ip(state, "rbl.ip", &module->ip) != 0)
  {
    if(state->char2addr("127.0.0.2", 0, &module->ip) != 0) CHAR2ADDR_FAILED;
  }

  count = state->config_maxn(state, "rbl.range");
  if((module->rbl_ip = (rbl_ip_t*)calloc((size_t)count, sizeof(module->rbl_ip[0]))) == NULL) CALLOC_FAILED;
  for(module->rbl_ip_count = 0; module->rbl_ip_count < count; module->rbl_ip_count++)
  {
    memset(&module->rbl_ip[module->rbl_ip_count], 0, sizeof(module->rbl_ip[0]));
    if(state->buffer_set(&buffer, "rbl.range.%d", module->rbl_ip_count + 1) != 0) BUFFER_SET_FAILED;
    if(state->config_get_ip_range(state, (char*)buffer.value, &module->rbl_ip[module->rbl_ip_count].range) != 0) CONFIG_GET_IP_RANGE_FAILED;
    module->rbl_ip[module->rbl_ip_count].family = (int)module->rbl_ip[module->rbl_ip_count].range.begin.a.sockaddr.sa_family;
  }

  count = state->config_maxn(state, "rbl.pattern");
  if((module->rbl_pattern = (char**)calloc((size_t)count, sizeof(module->rbl_pattern[0]))) == NULL) CALLOC_FAILED;
  for(module->rbl_pattern_count = 0; module->rbl_pattern_count < count; module->rbl_pattern_count++)
  {
    if(state->buffer_set(&buffer, "rbl.pattern.%d", module->rbl_pattern_count + 1) != 0) BUFFER_SET_FAILED;
    if((module->rbl_pattern[module->rbl_pattern_count] = state->config_get_lower(state, (char*)buffer.value)) == NULL) CONFIG_GET_LOWER_FAILED;
  }
  state->buffer_free(&buffer);

  if(module->rbl_ip_count == 0 && module->rbl_pattern_count == 0)
  {
    ERROR_IF(module->trace) (void)state->lprintf(TRACE, "No ranges or patterns are set");
    term(module_index, state);
    return -1;
  }
  state->config_free_node(state, "rbl");
  return 0;
}

static answer_t answer(int module_index, client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int lazy)
{
  buffer_t ptrname;
  address_t a;
  int i, j;
  char address[IPCHAR_SIZE];
  client_t *local_client = NULL;
  state_t *state = client->state;
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  address[0] = '\0';

  if(state->is_subdomain(module->rbl_domain, (char*)name->value) == 0) return ANSWER_UNSATISFIED;
  if(lazy != 0) return ANSWER_BLOCKED;
  if(dclass != CLASS_IN && dclass != CLASS_ANY) return ANSWER_BLOCKED;
  if(client->header.flags.ra == 0) return ANSWER_BLOCKED;

  if(type == TYPE_A)
  {
    if(state->addr_prefix(name, &a) == 0) return ANSWER_BLOCKED;
  }
  else if(type == TYPE_AAAA)
  {
    if(state->addr6_prefix(name, &a) == 0) return ANSWER_BLOCKED;
  }
  else
  {
    return ANSWER_BLOCKED;
  }

  for(i = 0; i < module->rbl_ip_count; i++)
  {
    if(state->is_in_range(&a, &module->rbl_ip[i].range) != 0)
    {
      rr_t *rr;
      if((type == TYPE_A && module->rbl_ip[i].family == AF_INET) || (type == TYPE_AAAA && module->rbl_ip[i].family == AF_INET6))
      {
        if((rr = state->rr_create(name, type, dclass, module->ttl, RCODE_NO_ERROR)) != NULL)
        {
          memcpy(&rr->p.address, &module->ip, sizeof(rr->p.address));
          INFO_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %s (matches on range)", (char*)name->value, state->class_name(dclass), state->type_name(type), state->addr2char(address, &rr->p.address));
          if(state->add_rr(client, rr_section, &rr) == 0)
          {
            state->free_rr(&rr);
            return ANSWER_SATISFIED;
          }
          state->free_rr(&rr);
        }
        break;
      }
    }
  }

  buffer_init(&ptrname);
  if(state->ptr_name(&ptrname, &a) != 0) return ANSWER_BLOCKED;

  if(state->init_client(state, &local_client) != 0)
  {
    state->free_client(local_client);
    state->buffer_free(&ptrname);
    return ANSWER_BLOCKED;
  }
  local_client->header.flags = client->header.flags;

  if(state->answer_question(local_client, RR_AN, &ptrname, TYPE_PTR, CLASS_IN, 0) != ANSWER_SATISFIED)
  {
    state->free_client(local_client);
    state->buffer_free(&ptrname);
    return ANSWER_BLOCKED;
  }

  for(i = 0; i < (int)local_client->header.ancount; i++)
  {
    if(local_client->answers[i]->rtype == TYPE_PTR && local_client->answers[i]->rclass == CLASS_IN && state->buffer_cmp(&local_client->answers[i]->name.name, &ptrname) == 0)
    {
      for(j = 0; j < module->rbl_pattern_count; j++)
      {
        if(state->match(module->rbl_pattern[j], (char*)local_client->answers[i]->p.ptr.name.value) != 0)
        {
          rr_t *rr;
          if((rr = state->rr_create(name, type, dclass, module->ttl, RCODE_NO_ERROR)) != NULL)
          {
            memcpy(&rr->p.address, &module->ip, sizeof(rr->p.address));
            INFO_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %s (matches on %s)", (char*)name->value, state->class_name(dclass), state->type_name(type),
                          state->addr2char(address, &rr->p.address), module->rbl_pattern[j]);
            if(state->add_rr(client, rr_section, &rr) == 0)
            {
              state->free_rr(&rr);
              state->buffer_free(&ptrname);
              state->free_client(local_client);
              return ANSWER_SATISFIED;
            }
            state->free_rr(&rr);
          }
        }
      }
    }
  }
  state->free_client(local_client);
  state->buffer_free(&ptrname);

  client->header.flags.rcode = RCODE_NAME_ERROR;
  return ANSWER_BLOCKED;
}

module_t unnamed_module = { "rbl", NULL, init, answer, NULL, term };
